import os.path
import sys

import mkdocs
from mkdocs.config import config_options
import mkdocs.plugins
from mkdocs.structure.files import Files
from mkdocs.utils import copy_file

from .exporter import ConfigurationError, DrawIoExporter, Configuration


log = mkdocs.plugins.log.getChild('drawio-exporter')


class DrawIoExporterPlugin(mkdocs.plugins.BasePlugin):
    """Draw.io Exporter MkDocs plugin.

    Contains only the bindings to MkDocs events, as it's difficult to isolate
    this module from MkDocs for testing.
    """

    config_scheme = (
        ('cache_dir', config_options.Type(str)),
        ('drawio_executable', config_options.Type(str)),
        ('drawio_args', config_options.Type(list, default=[])),
        ('format', config_options.Type(str, default='svg')),
        ('embed_format', config_options.Type(str, default='{img_open}{img_src}{img_close}')),
        ('sources', config_options.Type(str, default='*.drawio')),
    )

    exporter = None

    sources = []

    def on_config(self, config):
        self.exporter = DrawIoExporter(log, config['docs_dir'])

        self.config['cache_dir'] = self.exporter.prepare_cache_dir(
                self.config['cache_dir'])
        try:
            self.config['drawio_executable'] = self.exporter.prepare_drawio_executable(
                    self.config['drawio_executable'],
                    DrawIoExporter.DRAWIO_EXECUTABLE_NAMES,
                    self.exporter.drawio_executable_paths(sys.platform))
            self.exporter.validate_config(self.config)
        except ConfigurationError as e:
            raise mkdocs.exceptions.ConfigurationError(str(e))

        os.makedirs(self.config['cache_dir'], exist_ok=True)

        log.debug(f'Using Draw.io executable "{self.config["drawio_executable"]}", '
                f'arguments {self.config["drawio_args"]} and '
                f'cache directory "{self.config["cache_dir"]}"')

    def on_post_page(self, output_content, page, **kwargs):
        output_content, content_sources = self.exporter.rewrite_image_embeds(
                page.file.dest_path, output_content, self.config)

        self.sources += content_sources

        return output_content

    def on_files(self, files, config):
        keep = self.exporter.filter_cache_files(files, self.config['cache_dir'])
        log.debug(f'{len(keep)} files left after excluding cache')

        return Files(keep)

    def on_post_build(self, config):
        sources = set(self.sources)
        log.debug(f'Found {len(sources)} unique sources in {len(self.sources)} total embeds')
        self.sources = []

        for source in sources:
            dest_rel_path = f'{source.source_rel}-{source.page_index}.{self.config["format"]}'
            abs_src_path = os.path.join(config['docs_dir'], source.source_rel)
            abs_dest_path = os.path.join(config['site_dir'], dest_rel_path)
            cache_filename, exit_status = self.exporter.ensure_file_cached(
                    abs_src_path, source.source_rel, source.page_index,
                    self.config)

            if exit_status not in (None, 0):
                log.error(f'Export failed with exit status {exit_status}; skipping copy')
                continue

            try:
                copy_file(cache_filename, abs_dest_path)
            except FileNotFoundError:
                log.warning('Export successful, but wrote no output file')



import unittest
from unittest.mock import MagicMock, patch

import logging
import os
from os.path import join, sep

from ..exporter import Configuration, ConfigurationError, DrawIoExporter


class FileMock:
    src_path = None
    abs_src_path = None

    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)


class ExporterTests(unittest.TestCase):
    log = None

    def setUp(self):
        self.log = logging.getLogger(__name__)

    def make_exporter(self, docs_dir=None):
        if not docs_dir:
            docs_dir = sep + 'docs'
        return DrawIoExporter(self.log, docs_dir)

    def make_config(self, **kwargs):
        defaults = {
            'cache_dir': 'drawio-exporter',
            'drawio_executable': 'drawio',
            'drawio_args': [],
            'format': 'svg',
            'embed_format': '{img_open}{img_src}{img_close}',
            'sources': '*.drawio',
        }
        # FIXME: when dropping support for Python 3.8, replace with the merge
        # operator (|).
        values = {**defaults, **kwargs}
        return Configuration(**values)

    def test_drawio_executable_paths_warns_on_unknown_platform(self):
        self.log.warning = MagicMock()
        exporter = self.make_exporter()
        exporter.drawio_executable_paths('win32-but-stable')
        self.log.warning.assert_called_once()

    def test_prepare_cache_dir_defaults(self):
        exporter = self.make_exporter()
        assert len(exporter.prepare_cache_dir(sep + 'docs'))

    def test_prepare_cache_dir_resolves_relative_path(self):
        exporter = self.make_exporter()

        result = exporter.prepare_cache_dir('temp')
        assert os.path.isabs(result)
        assert result.startswith(sep + 'docs' + sep)

    def test_prepare_drawio_executable_aborts_on_missing_executable(self):
        exporter = self.make_exporter()
        with self.assertRaises(ConfigurationError):
            exporter.prepare_drawio_executable(
                    sep + join('does', 'not', 'exist'), [], [])

    def test_prepare_drawio_executable_uses_valid_specified_exectuable(self):
        exporter = self.make_exporter()
        # We need something that exists, and right now we're not verifying that
        # it's executable. This will do.
        expected = os.path.abspath(__file__)
        actual = exporter.prepare_drawio_executable(
                expected, exporter.DRAWIO_EXECUTABLE_NAMES, [])
        assert expected == actual

    @patch('shutil.which')
    @patch.dict(os.environ, {'PATH': sep + join('does', 'not', 'exist')})
    def test_prepare_drawio_executable_on_path(self, mock_which):
        exporter = self.make_exporter()
        mock_which.return_value = sep + join('does', 'not', 'exist', 'drawio')
        result = exporter.prepare_drawio_executable(None, exporter.DRAWIO_EXECUTABLE_NAMES, [])
        assert result == mock_which.return_value

    @patch('os.path.isfile')
    def test_prepare_drawio_executable_platform_specific(self, mock_isfile):
        expect = sep + join('does', 'not', 'exist', 'someotherdrawio')
        mock_isfile.side_effect = lambda p: p == expect

        exporter = self.make_exporter()

        result = exporter.prepare_drawio_executable(None, [], [
            sep + join('does', 'not', 'exist', 'drawio'),
            expect,
        ])
        assert result == expect

    def test_prepare_drawio_executable_raises_on_failure(self):
        exporter = self.make_exporter()

        with self.assertRaises(ConfigurationError):
            exporter.prepare_drawio_executable(None, [], [])

    def test_rewrite_image_embeds(self):
        page_dest_path = "index.html"
        source = '''<h1>Example text</h1>
<img alt="Some text" src="../some-diagram.drawio" />'''
        object_embed_format = '<object type="image/svg+xml" data="{img_src}"></object>'

        exporter = self.make_exporter()

        config = self.make_config(sources='*.nomatch')
        output_content, sources = exporter.rewrite_image_embeds(
                page_dest_path, source, config)
        assert output_content == source
        assert sources == []

        config = self.make_config()
        output_content, sources = exporter.rewrite_image_embeds(
                page_dest_path, source, config)
        assert output_content != source
        assert 'src="../some-diagram.drawio-0.svg"' in output_content
        assert len(sources) == 1

        config = self.make_config(embed_format=object_embed_format)
        output_content, sources = exporter.rewrite_image_embeds(
                page_dest_path, source, config)
        assert output_content != source
        assert '<object type="image/svg+xml" data="../some-diagram.drawio-0.svg"></object>' in output_content
        assert len(sources) == 1

    def test_filter_cache_files(self):
        files = [
            sep + join('docs', 'drawio-exporter', '0000000000000000000000000000000000000000-0'),
            sep + join('docs', 'index.md'),
        ]
        files = [FileMock(**{'abs_src_path': f}) for f in files]

        exporter = self.make_exporter()

        result = exporter.filter_cache_files(files, sep + join('docs', 'nomatch'))
        assert len(result) == 2

        result = exporter.filter_cache_files(files, sep + join('docs', 'drawio-exporter'))
        assert len(result) == 1
        assert files[0] not in result

    def test_ensure_file_cached(self):
        source = sep + join('docs', 'diagram.drawio')
        source_rel = 'diagram.drawio'
        drawio_executable = sep + join('bin', 'drawio')
        cache_dir = sep + join('docs', 'drawio-exporter')

        exporter = self.make_exporter()
        config = self.make_config(
            cache_dir=cache_dir,
            drawio_executable=drawio_executable,
        )

        exporter.make_cache_filename = MagicMock()
        exporter.make_cache_filename.return_value = join(cache_dir, '0000000000000000000000000000000000000000-0')

        exporter.export_file = MagicMock()
        exporter.export_file.return_value = 0

        cache_filename, exit_status = exporter.ensure_file_cached(
                source, source_rel, 0, config)
        assert cache_filename == exporter.make_cache_filename.return_value
        assert exit_status == 0

    def test_ensure_file_cached_aborts_if_drawio_executable_unavailable(self):
        source = sep + join('docs', 'diagram.drawio')
        source_rel = 'diagram.drawio'
        drawio_executable = None
        cache_dir = sep + join('docs', 'drawio-exporter')

        exporter = self.make_exporter()
        config = self.make_config(
            cache_dir=cache_dir,
            drawio_executable=drawio_executable,
        )

        exporter.export_file = MagicMock()
        exporter.export_file.return_value = 0

        self.log.warning = MagicMock()

        _, exit_status = exporter.ensure_file_cached(
                source, source_rel, 0, config)

        assert exit_status == None
        self.log.warning.assert_called_once()

    def test_ensure_file_cached_skips_export_if_cache_fresh(self):
        source = sep + join('docs', 'diagram.drawio')
        source_rel = 'diagram.drawio'
        drawio_executable = sep + join('bin', 'drawio')
        cache_dir = sep + join('docs', 'drawio-exporter')

        exporter = self.make_exporter()
        config = self.make_config(
            cache_dir=cache_dir,
            drawio_executable=drawio_executable,
        )

        exporter.make_cache_filename = MagicMock()
        exporter.make_cache_filename.return_value = join(cache_dir, '0000000000000000000000000000000000000000-0')

        exporter.use_cached_file = MagicMock()
        exporter.use_cached_file.return_value = True

        exporter.export_file = MagicMock()
        exporter.export_file.return_value = 0

        cache_filename, exit_status = exporter.ensure_file_cached(
                source, source_rel, 0, config)

        assert cache_filename == exporter.make_cache_filename.return_value
        assert exit_status == None
        exporter.use_cached_file.assert_called_once()
        assert not exporter.export_file.called

    def test_ensure_file_cached_returns_exit_status_if_non_zero(self):
        source = sep + join('docs', 'diagram.drawio')
        source_rel = 'diagram.drawio'
        drawio_executable = sep + join('bin', 'drawio')
        cache_dir = sep + join('docs', 'drawio-exporter')

        exporter = self.make_exporter()
        config = self.make_config(
            cache_dir=cache_dir,
            drawio_executable=drawio_executable,
        )

        exporter.make_cache_filename = MagicMock()
        exporter.make_cache_filename.return_value = join(cache_dir, '0000000000000000000000000000000000000000-0')

        exporter.use_cached_file = MagicMock()
        exporter.use_cached_file.return_value = False

        exporter.export_file = MagicMock()
        exporter.export_file.return_value = 1

        self.log.error = MagicMock()

        cache_filename, exit_status = exporter.ensure_file_cached(
                source, source_rel, 0, config)

        assert exit_status == 1

    def test_make_cache_filename(self):
        cache_dir = sep + 'docs'

        exporter = self.make_exporter()

        results = [
            exporter.make_cache_filename('diagram.drawio', 0, cache_dir),
            exporter.make_cache_filename('other-diagram.drawio', 0, cache_dir),
            exporter.make_cache_filename('other-diagram.drawio', 1, cache_dir),
        ]

        for result in results:
            assert result.startswith(cache_dir)
        assert len(set(results)) == 3

    @patch('os.path.exists')
    @patch('os.path.getmtime')
    def test_use_cached_file(self, getmtime_mock, exists_mock):
        cache_filename = sep + join('docs', 'drawio-exporter', '0000000000000000000000000000000000000000-0')
        source = sep + join('docs', 'diagram.drawio')

        exists_mock.return_value = True

        exporter = self.make_exporter()

        def getmtime_return_value(path):
            if path == cache_filename:
                return 1577043612.1444612
            elif path == source:
                return 1577133635.3318102
            else:
                raise ValueError(f'didn\'t expect path "{path}"')
        getmtime_mock.side_effect = getmtime_return_value

        result = exporter.use_cached_file(source, cache_filename)
        assert result == False

    @patch('subprocess.call')
    def test_export_file(self, call_mock):
        source = sep + join('docs', 'diagram.drawio')
        dest = sep + join('docs', 'diagram.drawio-0.svg')
        drawio_executable = sep + join('bin', 'drawio')

        exporter = self.make_exporter()
        config = self.make_config(
            drawio_executable=drawio_executable,
        )

        call_mock.return_value = 0

        result = exporter.export_file(
                source, 0, dest, config)

        assert result == 0
        call_mock.assert_called_once()

    @patch('subprocess.call')
    def test_export_file_logs_exc_on_raise(self, call_mock):
        source = sep + join('docs', 'diagram.drawio')
        dest = sep + join('docs', 'diagram.drawio-0.svg')
        drawio_executable = sep + join('bin', 'drawio')

        exporter = self.make_exporter()
        config = self.make_config(
            drawio_executable=drawio_executable,
        )

        call_mock.side_effect = OSError()

        self.log.exception = MagicMock()

        result = exporter.export_file(
                source, 0, dest, config)

        assert result == None
        self.log.exception.assert_called_once()
        call_mock.assert_called_once()

    @patch('subprocess.call')
    def test_export_file_honours_drawio_args(self, call_mock):
        source = sep + join('docs', 'diagram.drawio')
        page_index = 0
        dest = sep + join('docs', 'diagram.drawio-0.svg')
        drawio_executable = sep + join('bin', 'drawio')

        exporter = self.make_exporter()
        config = self.make_config(
            drawio_executable=drawio_executable,
            format='svg'
        )

        def test_drawio_args(config: Configuration, drawio_args):
            test_config = {**config, 'drawio_args': drawio_args}
            exporter.export_file(
                    source, page_index, dest, test_config)
            call_mock.assert_called_with([
                config['drawio_executable'],
                '--export', source,
                '--page-index', str(page_index),
                '--output', dest,
                '--format', config['format'],
            ] + drawio_args)

        test_drawio_args(config, [])
        test_drawio_args(config, ['--no-sandbox'])


if __name__ == '__main__':
    unittest.main()



# FIXME: drop once we upgrade to Python 3.9
from __future__ import annotations

import fnmatch
import hashlib
import os.path
import re
import shutil
import subprocess
import sys
from typing import TypedDict
import urllib.parse


IMAGE_RE = re.compile('(<img[^>]+src=")([^">]+)("\\s*\\/?>)')


class Configuration(TypedDict):
    """Draw.io Exporter MkDocs plugin configuration.

    Contains the resolved configuration values, including defaults and any
    computed values (such as paths to binaries).

    Seems ugly to shadow BasePlugin.config_scheme, but improves type hints and
    allows us to more easily pass configuration around."""

    cache_dir: str
    drawio_executable: str
    drawio_args: list[str]
    format: str
    embed_format: str
    sources: str


class ConfigurationError(Exception):
    """Configuration exception.

    Used to signal that MkDocs should exit, as the plugin configuration is
    invalid and we'll be unable to export diagrams.
    """

    key = None
    """Configuration key.

    :type str:"""

    value = None
    """Configuration value.

    :type mixed:"""

    message = None
    """Explanatory message describing the problem.

    :type str:"""

    def __init__(self, key, value, message):
        """Initialise a ConfigurationError.

        :param str key: Configuration key.
        :param mixed value: Configuration value.
        :param str message: Explanatory message describing the problem.
        """
        self.key = key
        self.value = value
        self.message = message
        Exception.__init__(self, self.message)

    def __str__(self):
        return f'drawio-exporter: value "{self.value}" for key "{self.key}" is invalid: {self.message}'

    def drawio_executable(value, message):
        """Raise an error for a misconfigured Draw.io path.

        :param str value: Configured Draw.io executable path.
        :param str message: Explanatory message describing the problem.
        """
        return ConfigurationError('drawio_executable', value, message)


class Source:
    """Diagram source.

    Sources are pairs of filenames and page indices which can be exported to
    produce a static image. The relative path of the source within the
    documentation directory must be resolved after instantiation due to MkDocs's
    design.
    """

    source_embed = None
    """Path of the embedded resource, relative to parent page.

    :type: str
    """

    source_rel = None
    """Path of the source, relative to the documentation directory.

    :type: str
    """

    page_index = None
    """Page index within the document.

    :type: int"""

    def __init__(self, source_embed, page_index):
        """Initialise a Source.

        :param str source_embed: Path of the embedded resource.
        :param int page_index: Page index within the document.
        """
        self.source_embed = source_embed
        self.page_index = page_index

    def __eq__(self, other):
        return self.source_rel == other.source_rel \
                and self.page_index == other.page_index

    def __hash__(self):
        return hash((
            'source_rel', self.source_rel,
            'page_index', self.page_index,
        ))

    def __repr__(self):
        return f"Source({self.source_embed}, {self.page_index}, {self.source_rel})"

    def resolve_rel_path(self, page_dest_path):
        """Resolve the path of the source, relative to the documentation directory.

        :param str page_dest_path: The destination path of the parent page.
        """
        unescaped_source_embed = urllib.parse.unquote(self.source_embed)
        self.source_rel = os.path.normpath(os.path.join(
                os.path.dirname(page_dest_path),
                unescaped_source_embed))


class DrawIoExporter:
    """Draw.io Exporter.

    The logic for the export process lives here. The bindings to the MkDocs
    plugin events is kept separate to ease testing.
    """

    log = None
    """Log.

    :type: logging.Logger
    """

    docs_dir = None
    """Draw.io docs_dir.

    :type str:
    """

    def __init__(self, log, docs_dir):
        """Initialise.

        :param logging.Logger log: Where to log.
        :param str docs_dir: MkDocs docs_dir.
        """
        self.log = log
        self.docs_dir = docs_dir

    DRAWIO_EXECUTABLE_NAMES = ['drawio', 'draw.io']
    """Draw.io executable names."""

    def drawio_executable_paths(self, platform):
        """Get the Draw.io executable paths for the platform.

        Declared as a function to allow us to use API/environment information
        available only when running under the specified platform.

        :param str platform: sys.platform.
        :return list(str): All known paths.
        """
        if platform == 'darwin':
            applications = [
                os.path.expanduser('~/Applications'),
                '/Applications',
            ]
            drawio_path = os.path.join('draw.io.app', 'Contents', 'MacOS', 'draw.io')
            return [os.path.join(dir, drawio_path) for dir in applications]
        elif platform.startswith('linux'):
            return ['/opt/draw.io/drawio']
        elif platform == 'win32':
            program_files = [os.environ['ProgramFiles']]
            if 'ProgramFiles(x86)' in os.environ:
                program_files.append(os.environ['ProgramFiles(x86)'])
            return [os.path.join(dir, 'draw.io', 'draw.io.exe') for dir in program_files]
        else:
            self.log.warning(f'Draw.io executable paths not known for platform "{platform}"')

    def prepare_cache_dir(self, cache_dir):
        """Ensure the cache path is set, absolute and exists.

        :param str cache_dir: Configured cache directory.
        :param str docs_dir: Docs directory, in which to base relative cache directories.
        :return str: Final cache directory.
        """
        if not cache_dir:
           cache_dir = 'drawio-exporter'
        if not os.path.isabs(cache_dir):
            cache_dir = os.path.join(self.docs_dir, cache_dir)
        return cache_dir

    def prepare_drawio_executable(self, executable, executable_names, platform_executable_paths):
        """Ensure the Draw.io executable path is configured, or guess it.

        :param str executable: Configured Draw.io executable.
        :param list(str) executable_names: Candidate executable names to seek on PATH.
        :param list(str) platform_executable_paths: Candidate platform-specific executable paths.
        :return str: Final Draw.io executable.
        """
        if executable:
            if not os.path.isfile(executable):
                raise ConfigurationError.drawio_executable(
                        executable, "executable didn't exist")
            return executable

        for name in executable_names:
            executable = shutil.which(name)
            if executable:
                self.log.debug(f'Found Draw.io executable "{name}" at "{executable}"')
                return executable

        candidates = platform_executable_paths
        self.log.debug(f'Trying paths {candidates} for platform "{sys.platform}"')
        for candidate in candidates:
            if os.path.isfile(candidate):
                self.log.debug(f'Found Draw.io executable for platform "{sys.platform}" at "{candidate}"')
                return candidate

        raise ConfigurationError.drawio_executable(
                None, 'Unable to find Draw.io executable; ensure it\'s on PATH or set drawio_executable option')

    def validate_config(self, config: Configuration):
        """Validate the configuration.

        :param dict config: Configuration.
        :return bool: True if configuration is valid.
        """
        if '{content}' in config['embed_format'] and config['format'] != 'svg':
            raise ConfigurationError(
                    'embed_format', config['embed_format'],
                    'cannot inline content of non-SVG format')

    def rewrite_image_embeds(self, page_dest_path, output_content, config: Configuration):
        """Rewrite image embeds.

        :param str page_dest_path: Destination path.
        :param str output_content: Content to rewrite.
        :param str sources: Glob to match Draw.io diagram filenames.
        :param str format: Desired export format.
        :param str embed_format: Format string to rewrite <img> tags with.
        :return str: Rewritten content.
        """
        content_sources = []

        def replace(match):
            try:
                filename, page_index = match.group(2).rsplit('#', 1)
            except ValueError:
                filename = match.group(2)
                page_index = 0

            if fnmatch.fnmatch(filename, config["sources"]):
                source = Source(filename, page_index)
                source.resolve_rel_path(page_dest_path)
                content_sources.append(source)
                img_src = f"{filename}-{page_index}.{config['format']}"

                # Cache the file on-demand and read file content only if we
                # need to inline the file's content.
                content = None
                if "{content}" in config["embed_format"]:
                    img_path = self.make_cache_filename(
                            source.source_rel, page_index, config['cache_dir'])

                    abs_src_path = os.path.join(self.docs_dir, source.source_rel)
                    _, exit_status = self.ensure_file_cached(
                            abs_src_path, source.source_rel, source.page_index,
                            config)

                    if exit_status not in (None, 0):
                        self.log.error(f'Export failed with exit status {exit_status}; skipping rewrite')
                        return match.group(0)

                    with open(img_path, "r") as f:
                        content = f.read()

                return config["embed_format"].format(
                        img_open=match.group(1), img_close=match.group(3),
                        img_src=img_src, content=content)
            else:
                return match.group(0)
        output_content = IMAGE_RE.sub(replace, output_content)

        return (output_content, content_sources)

    def filter_cache_files(self, files, cache_dir):
        """Remove cache files from the generated output.

        :param list(mkdocs.structure.File): Files to filter.
        :param str cache_dir: Cache directory.
        :return list(mkdocs.structure.File): Filtered files.
        """
        return [f for f in files if not f.abs_src_path.startswith(cache_dir)]

    def ensure_file_cached(self, source, source_rel, page_index, config: Configuration):
        """Ensure cached copy of output exists.

        :param str source: Source path, absolute.
        :param str source_rel: Source path, relative to docs directory.
        :param int page_index: Page index, numbered from zero.
        :param str drawio_executable: Path to the configured Draw.io executable.
        :param list(str) drawio_args: Additional arguments to append to the Draw.io export command.
        :param str cache_dir: Export cache directory.
        :param str format: Desired export format.
        :return tuple(str, int): Cached export filename.
        """
        cache_filename = self.make_cache_filename(source_rel, page_index, config['cache_dir'])
        exit_status = None

        if self.use_cached_file(source, cache_filename):
            self.log.debug(f'Source file appears unchanged; using cached copy from "{cache_filename}"')
        else:
            if not config['drawio_executable']:
                self.log.warning(f'Skipping export of "{source}" as Draw.io executable not available')
                return (None, exit_status)

            self.log.debug(f'Exporting "{source}" to "{cache_filename}"')
            exit_status = self.export_file(
                    source, page_index, cache_filename, config)

        return (cache_filename, exit_status)

    def make_cache_filename(self, source, page_index, cache_dir):
        """Make the cached filename.

        :param str source: Source path, relative to the docs directory.
        :param int page_index: Page index, numbered from zero.
        :param str cache_dir: Export cache directory.
        :return str: Resulting filename.
        """
        filename_hash = hashlib.sha1(source.encode('utf-8')).hexdigest()
        basename = f'{filename_hash}-{page_index}'
        return os.path.join(cache_dir, basename)

    def use_cached_file(self, source, cache_filename):
        """Is the cached copy up to date?

        :param str source: Source path, relative to the docs directory.
        :param str cache_filename: Export cache filename.
        :return bool: True if cache is up to date, else False.
        """
        return os.path.exists(cache_filename) \
                and os.path.getmtime(cache_filename) >= os.path.getmtime(source)

    def export_file(self, source, page_index, dest, config: Configuration):
        """Export an individual file.

        :param str source: Source path, absolute.
        :param int page_index: Page index, numbered from zero.
        :param str dest: Destination path, within cache.
        :param str drawio_executable: Path to the configured Draw.io executable.
        :param list(str) drawio_args: Additional arguments to append to the Draw.io export command.
        :param str format: Desired export format.
        :return int: The Draw.io exit status.
        """
        cmd = [
            config['drawio_executable'],
            '--export', source,
            '--page-index', str(page_index),
            '--output', dest,
            '--format', config['format'],
        ]
        cmd += config['drawio_args']

        try:
            self.log.debug(f'Using export command {cmd}')
            return subprocess.call(cmd)
        except:
            self.log.exception('Subprocess raised exception')
